// **************************************************************************************
// el Comandante                                                    guerilla.bplaced.net
// --------------------------------------------------------------------------------------
// Information:
// Receiving a 32 bit (uint32_t) value from a ESPNow transmitter using the esp32 buildin
// wireless ESPNow protocol. This esp32 receiver can check the received uint32_t bits and
// perform GPIO action, i.e. remotly control motors, LED's, up to 32 devices...
// Check the serial monitor for the MAC of this device, insert it in the transmitter software
// --------------------------------------------------------------------------------------
// Used Libraries:
// TFT_eSPI                 --> https://github.com/Bodmer/TFT_eSPI
// **************************************************************************************

#include <Arduino.h>                            // Arduino
#include <esp_now.h>                            // ESP-NOW
#include <WiFi.h>                               // WiFi
#include <TFT_eSPI.h>                           // TFT_espi library
TFT_eSPI tft = TFT_eSPI();

// ######################################################################################
// Change values in this area to match your envoirement and desires
// --------------------------------------------------------------------------------------
// notting to change
// ######################################################################################

// espNOW - sender data structure, must match the receiver data structure, 32 bit aligned!
typedef struct struct_espNowData
{uint32_t espNowData32Bit = 0;}
struct_espNowData;

// espNOW - create a struct_espNowData called espNowData
struct_espNowData espNowData;

// Vars
bool      espNowMessageNew = false;
uint32_t  i;
uint32_t  x;
uint32_t  z;
uint16_t  tftWidth;
uint16_t  tftHeight;
uint16_t  fontWidth;
uint16_t  fontHeight;
uint16_t  tftBaseX = 4;
uint16_t  tftBaseY = 8;
uint16_t  tftSpaceX = 0;
uint16_t  tftSpaceY = 4;
uint8_t   espNowData32BitArray[32];

// screen elements position...
uint16_t  espNowTitleTextXpos;
uint16_t  espNowTitleTextYpos;
uint16_t  espNowTitleDataXpos;
uint16_t  espNowTitleDataYpos;

uint16_t  espNowTitleLineXpos;
uint16_t  espNowTitleLineYpos;

uint16_t  espNowDatasBitsXpos;
uint16_t  espNowDatasBitsYpos;

uint16_t  espNowByte0BitsXpos[8];
uint16_t  espNowByte0BitsYpos[8];
uint16_t  espNowByte1BitsXpos[8];
uint16_t  espNowByte1BitsYpos[8];
uint16_t  espNowByte2BitsXpos[8];
uint16_t  espNowByte2BitsYpos[8];
uint16_t  espNowByte3BitsXpos[8];
uint16_t  espNowByte3BitsYpos[8];

uint16_t  espNowStatusLineXpos;
uint16_t  espNowStatusLineYpos;

char      espNowStatusTextChar[2][9]  = {"Init", "Receive"};
uint16_t  espNowStatusTextXpos[2];
uint16_t  espNowStatusTextYpos[2];
uint16_t  espNowStatusTextWidth[2];
uint16_t  espNowStatusIconXpos[2];
uint16_t  espNowStatusIconYpos[2];
uint16_t  espNowStatusIconRadius[2];


// **************************************************************************************
// espNOW - callback function
void espNowDataReceive(const uint8_t * mac, const uint8_t *incomingData, int len)
{
  if (len == sizeof(espNowData))
  {
    memcpy(&espNowData, incomingData, sizeof(espNowData));
    tft.fillCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0x0700);
    tft.drawCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0x9492);
    //x = espNowData.espNowData32Bit;
    espNowMessageNew = true;
  }
  else
  {
    memset(&espNowData.espNowData32Bit, 0, sizeof(espNowData));    
    tft.fillCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0xF800);
    tft.drawCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0x9492);
    //x = 0;
    espNowMessageNew = false;
  }
}


// **************************************************************************************
// Setup
void setup(void)
{
  pinMode(TFT_BL, OUTPUT);                                // GPIO - init pin
  digitalWrite(TFT_BL, 0);                                // Switch backlight off
  Serial.begin(115200);                                   // serial monitor
  tft.init();                                             // TFT_espi - init
  tft.setRotation(3);
  ScreenSize();
  ScreenBuild();
  digitalWrite(TFT_BL, 1);                                // switch backlight on
  ESPNowSetup();
  Serial.println(WiFi.macAddress());                      // prints the MAC of this decice
}


// **************************************************************************************
// Main loop
void loop()
{
  if (espNowMessageNew == true)
  {
    x = espNowData.espNowData32Bit;
    BitsArray();
    ScreenUpdate();
    espNowMessageNew = false;
  }
}


// **************************************************************************************
// Screen - size
void ScreenSize()
{
  tftWidth = tft.width();                                 // get tft width
  tftHeight = tft.height();                               // get tft height
  tft.setTextFont(1);
  if (tftWidth >= 320 || tftHeight >= 320)                // big screen (i.e. WT32-SC01 - 480 x 320)
  {
    tft.setTextSize(2);
    fontWidth = tft.textWidth("0", 1);                    // get the text width
    fontHeight = tft.fontHeight(1);                       // get the font height
    tftSpaceX = fontWidth / 2;
    ElementsArray();
  }
  else if (tftWidth < 320 || tftHeight < 320)             // small screen (i.e. LILYGO TTGO-T-Display - 240 x 135)
  {
    tft.setTextSize(1);
    fontWidth = tft.textWidth("0", 1);                    // get the text width
    fontHeight = tft.fontHeight(1);                       // get the font height
    tftSpaceX = fontWidth / 2;
    ElementsArray();    
  }
}


// **************************************************************************************
// Screen - fill elements array with Xpos / Ypos
void ElementsArray()
{
  espNowTitleTextXpos = tftBaseX;
  espNowTitleTextYpos = tftBaseY;
  espNowTitleDataXpos = tftWidth / 2;
  espNowTitleDataYpos = tftBaseY;
  espNowTitleLineXpos = 0;
  espNowTitleLineYpos = tftBaseY + (fontHeight + tftSpaceY) * 1;
  espNowDatasBitsXpos = tftBaseX;
  espNowDatasBitsYpos = tftBaseY + (fontHeight + tftSpaceY) * 1.75;

  for (i = 0; i < 8; i++)
  {
  espNowByte0BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte0BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 3;
  espNowByte1BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte1BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 4;
  espNowByte2BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte2BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 5;
  espNowByte3BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte3BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 6;
  }

  espNowStatusLineXpos = 0;
  espNowStatusLineYpos = tftBaseY + (fontHeight + tftSpaceY) * 7.5;

  espNowStatusIconRadius[0] = tftWidth / 100 * 2;
  espNowStatusIconRadius[1] = tftWidth / 100 * 2;
  espNowStatusTextWidth[0] = tft.textWidth(espNowStatusTextChar[0], 1);
  espNowStatusTextWidth[1] = tft.textWidth(espNowStatusTextChar[1], 1);
  espNowStatusTextXpos[0] = tftBaseX + tftWidth / 4 * 1;
  espNowStatusTextYpos[0] = tftBaseY + (fontHeight + tftSpaceY) * 8;
  espNowStatusTextXpos[1] = tftBaseX + tftWidth / 4 * 2.3;
  espNowStatusTextYpos[1] = tftBaseY + (fontHeight + tftSpaceY) * 8;
  espNowStatusIconXpos[0] = espNowStatusTextWidth[0] / 2 + tftWidth / 4 * 1;
  espNowStatusIconYpos[0] = tftBaseY + (fontHeight + tftSpaceY) * 9.5;
  espNowStatusIconXpos[1] = espNowStatusTextWidth[1] / 2 + tftWidth / 4 * 2.3;
  espNowStatusIconYpos[1] = tftBaseY + (fontHeight + tftSpaceY) * 9.5;
}

 
// **************************************************************************************
// Screen - build
void ScreenBuild()
{
  tft.fillScreen(0x0000);
  tft.setTextDatum(TL_DATUM);

  tft.setTextColor(0x0700, 0x0000);
  tft.drawString("Received Hex: ", espNowTitleTextXpos, espNowTitleTextYpos);

  tft.drawLine(espNowTitleLineXpos, espNowTitleLineYpos, tftWidth, espNowTitleLineYpos, 0x8410);

  tft.setTextColor(0x07FF, 0x0000);
  tft.drawString("Bits - High Byte First:", espNowDatasBitsXpos, espNowDatasBitsYpos);

  tft.drawLine(espNowStatusLineXpos, espNowStatusLineYpos, tftWidth, espNowStatusLineYpos, 0x8410);

  tft.setTextColor(0xFFFE, 0x0000);
  for (i = 0; i < 2; i++)
  {
    tft.drawString(espNowStatusTextChar[i], espNowStatusTextXpos[i], espNowStatusTextYpos[i]);
  }
}


// **************************************************************************************
// Screen - update
void ScreenUpdate()
{
  tft.setTextDatum(TL_DATUM);

  tft.setTextColor(0xFFFE, 0x0000);
  tft.setCursor(espNowTitleDataXpos, espNowTitleDataYpos);
  tft.printf("0x%.8X", x);

  tft.setTextColor(0xFFFE, 0x0000);
  for (i = 0; i < 8; i++)
  {
    tft.drawNumber(espNowData32BitArray[i], espNowByte0BitsXpos[i], espNowByte0BitsYpos[i]);
    tft.drawNumber(espNowData32BitArray[i + 8], espNowByte1BitsXpos[i], espNowByte1BitsYpos[i]);
    tft.drawNumber(espNowData32BitArray[i + 16], espNowByte2BitsXpos[i], espNowByte2BitsYpos[i]);
    tft.drawNumber(espNowData32BitArray[i + 24], espNowByte3BitsXpos[i], espNowByte3BitsYpos[i]);
  }
}


// **************************************************************************************
// Screen - create bit array from uint32_t to transmit
void BitsArray()
{
  z = 31;
  for (i = 0; i < 32; i++)
  {
    espNowData32BitArray[z] = 0 != (x & (1 << i));
    z--;
  }
}


// **************************************************************************************
void ESPNowSetup()
{
  // ESP-NOW - set device as a Wi-Fi Station
  WiFi.mode(WIFI_STA);

  // ESP-NOW - init
  if (esp_now_init() != ESP_OK)
  {
    tft.fillCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0xF800);
    tft.drawCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0x9492);
  }
  else
  {
    tft.fillCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0x0700);
    tft.drawCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0x9492);
  }

  // ESP-NOW - register for reveive CB to get the trasnmitted packet
  esp_now_register_recv_cb(espNowDataReceive);;
}
