// **************************************************************************************
// el Comandante                                                    guerilla.bplaced.net
// --------------------------------------------------------------------------------------
// Information:
// Transmitting a 32 bit (uint32_t) value to a ESPNow receiver using the esp32 buildin
// wireless ESPNow protocol. The esp32 receiver can check the received uint32_t bits and
// perform GPIO action, i.e. control motors, LED's, up to 32 devices...
// --------------------------------------------------------------------------------------
// Used Libraries:
// TFT_eSPI                 --> https://github.com/Bodmer/TFT_eSPI
// **************************************************************************************

#include <Arduino.h>                            // Arduino
#include <esp_now.h>                            // ESP-NOW
#include <WiFi.h>                               // WiFi
#include <TFT_eSPI.h>                           // TFT_espi library
TFT_eSPI tft = TFT_eSPI();

// ######################################################################################
// Change values in this area to match your envoirement and desires
// --------------------------------------------------------------------------------------
// espNOW Receiver MAC address - REPLACE WITH YOUR ESP32-NOW RECEIVER MAC Address
uint8_t   broadcastAddress[] = {0x78, 0x00, 0x00, 0x00, 0x00, 0xAC};
// Option to transmit random hex value, 0 = disable, 1 = enable
uint8_t flagRandom = 1;
// Option to transmit the same value if flagRandom = 0, always uncomment one option (1, 2)
// 1 - The 32 bits transmitting, bits position 1F ... 00, leading 0b is the binary operator
//uint32_t x = 0b11010111000101110101010101010101; // Hex = 0xD7175555
// 2 - The 32 bit hex value transmitting
uint32_t x = 0xD7175555;
// Delay in ms for sending new data
uint32_t espNowSendDelay = 10000;
// espNow message delay in ms, the esp32 needs time to proceed a message
uint8_t espNowMsgDelay = 50;
// ######################################################################################

// espNOW - sender data structure, must match the receiver data structure, 32 bit aligned!
typedef struct struct_espNowData
{uint32_t espNowData32Bit = 0;}
struct_espNowData;

// espNOW - create a struct_espNowData called espNowData
struct_espNowData espNowData;

// espNOW - create a variable esp_now_peer_info_t to store peer information
esp_now_peer_info_t peerInfo;

// Vars
uint32_t  update;
uint32_t  i;
uint32_t  z;
uint16_t  tftWidth;
uint16_t  tftHeight;
uint16_t  fontWidth;
uint16_t  fontHeight;
uint16_t  tftBaseX = 4;
uint16_t  tftBaseY = 8;
uint16_t  tftSpaceX = 0;
uint16_t  tftSpaceY = 4;
uint8_t   espNowData32BitArray[32];

// screen elements position...
uint16_t  espNowTitleTextXpos;
uint16_t  espNowTitleTextYpos;
uint16_t  espNowTitleDataXpos;
uint16_t  espNowTitleDataYpos;

uint16_t  espNowTitleLineXpos;
uint16_t  espNowTitleLineYpos;

uint16_t  espNowDatasBitsXpos;
uint16_t  espNowDatasBitsYpos;

uint16_t  espNowByte0BitsXpos[8];
uint16_t  espNowByte0BitsYpos[8];
uint16_t  espNowByte1BitsXpos[8];
uint16_t  espNowByte1BitsYpos[8];
uint16_t  espNowByte2BitsXpos[8];
uint16_t  espNowByte2BitsYpos[8];
uint16_t  espNowByte3BitsXpos[8];
uint16_t  espNowByte3BitsYpos[8];

uint16_t  espNowStatusLineXpos;
uint16_t  espNowStatusLineYpos;

char      espNowStatusTextChar[4][9] = {"Init", "Peer", "Send", "Delivery"};
uint16_t  espNowStatusTextXpos[4];
uint16_t  espNowStatusTextYpos[4];
uint16_t  espNowStatusTextWidth[4];
uint16_t  espNowStatusIconXpos[4];
uint16_t  espNowStatusIconYpos[4];
uint16_t  espNowStatusIconRadius[4];


// **************************************************************************************
// espNOW - callback function
void ESPNowOnSend(const uint8_t *mac_addr, esp_now_send_status_t status)
{
  if (status != ESP_NOW_SEND_SUCCESS)
  {
    tft.fillCircle(espNowStatusIconXpos[3], espNowStatusIconYpos[3], espNowStatusIconRadius[3], 0xF800);
    tft.drawCircle(espNowStatusIconXpos[3], espNowStatusIconYpos[3], espNowStatusIconRadius[3], 0x9492);
  }
  else
  {
    tft.fillCircle(espNowStatusIconXpos[3], espNowStatusIconYpos[3], espNowStatusIconRadius[3], 0x0700);
    tft.drawCircle(espNowStatusIconXpos[3], espNowStatusIconYpos[3], espNowStatusIconRadius[3], 0x9492);
  }
}


// **************************************************************************************
// Setup
void setup(void)
{
  pinMode(TFT_BL, OUTPUT);                                // GPIO - init pin
  digitalWrite(TFT_BL, 0);                                // Switch backlight off
  tft.init();                                             // TFT_espi - init
  tft.setRotation(3);
  ScreenSize();
  ScreenBuild();
  digitalWrite(TFT_BL, 1);                                // switch backlight on
  ESPNowSetup();
  update = millis();
}


// **************************************************************************************
// Main loop
void loop()
{
  if (update <= millis())
  {
    if (flagRandom == 1)
    {x = esp_random();}                                   // true random if WiFi/Bluetooth enabled
    ESPNowSend();
    BitsArray();
    ScreenUpdate();
    update = millis() + espNowSendDelay;
  }
}


// **************************************************************************************
// Screen - size
void ScreenSize()
{
  tftWidth = tft.width();                                 // get tft width
  tftHeight = tft.height();                               // get tft height
  tft.setTextFont(1);
  if (tftWidth >= 320 || tftHeight >= 320)                // big screen (i.e. WT32-SC01 - 480 x 320)
  {
    tft.setTextSize(2);
    fontWidth = tft.textWidth("0", 1);                    // get the text width
    fontHeight = tft.fontHeight(1);                       // get the font height
    tftSpaceX = fontWidth / 2;
    ElementsArray();
  }
  else if (tftWidth < 320 || tftHeight < 320)             // small screen (i.e. LILYGO TTGO-T-Display - 240 x 135)
  {
    tft.setTextSize(1);
    fontWidth = tft.textWidth("0", 1);                    // get the text width
    fontHeight = tft.fontHeight(1);                       // get the font height
    tftSpaceX = fontWidth / 2;
    ElementsArray();    
  }
}


// **************************************************************************************
// Screen - fill elements array with Xpos / Ypos
void ElementsArray()
{
  espNowTitleTextXpos = tftBaseX;
  espNowTitleTextYpos = tftBaseY;
  espNowTitleDataXpos = tftWidth / 2;
  espNowTitleDataYpos = tftBaseY;
  espNowTitleLineXpos = 0;
  espNowTitleLineYpos = tftBaseY + (fontHeight + tftSpaceY) * 1;
  espNowDatasBitsXpos = tftBaseX;
  espNowDatasBitsYpos = tftBaseY + (fontHeight + tftSpaceY) * 1.75;

  for (i = 0; i < 8; i++)
  {
  espNowByte0BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte0BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 3;
  espNowByte1BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte1BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 4;
  espNowByte2BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte2BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 5;
  espNowByte3BitsXpos[i] = tftBaseX + fontWidth * 2 * i;
  espNowByte3BitsYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 6;
  }

  espNowStatusLineXpos = 0;
  espNowStatusLineYpos = tftBaseY + (fontHeight + tftSpaceY) * 7.5;

  for (i = 0; i < 4; i++)
  {
    espNowStatusIconRadius[i] = tftWidth / 100 * 2;
    espNowStatusTextXpos[i] = tftBaseX + tftWidth / 4 * i;
    espNowStatusTextYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 8;
    espNowStatusTextWidth[i] = tft.textWidth(espNowStatusTextChar[i], 1);
    espNowStatusIconXpos[i] = espNowStatusTextWidth[i] / 2 + tftWidth / 4 * i;
    espNowStatusIconYpos[i] = tftBaseY + (fontHeight + tftSpaceY) * 9.5;
  }
}

 
// **************************************************************************************
// Screen - build
void ScreenBuild()
{
  tft.fillScreen(0x0000);
  tft.setTextDatum(TL_DATUM);

  tft.setTextColor(0x0700, 0x0000);
  tft.drawString("Transmitting Hex: ", espNowTitleTextXpos, espNowTitleTextYpos);

  tft.drawLine(espNowTitleLineXpos, espNowTitleLineYpos, tftWidth, espNowTitleLineYpos, 0x8410);

  tft.setTextColor(0x07FF, 0x0000);
  tft.drawString("Bits - High Byte First:", espNowDatasBitsXpos, espNowDatasBitsYpos);

  tft.drawLine(espNowStatusLineXpos, espNowStatusLineYpos, tftWidth, espNowStatusLineYpos, 0x8410);

  tft.setTextColor(0xFFFE, 0x0000);
  for (i = 0; i < 4; i++)
  {
    tft.drawString(espNowStatusTextChar[i], espNowStatusTextXpos[i], espNowStatusTextYpos[i]);
  }
}


// **************************************************************************************
// Screen - update
void ScreenUpdate()
{
  tft.setTextDatum(TL_DATUM);

  tft.setTextColor(0xFFFE, 0x0000);
  tft.setCursor(espNowTitleDataXpos, espNowTitleDataYpos);
  tft.printf("0x%.8X", x);

  tft.setTextColor(0xFFFE, 0x0000);
  for (i = 0; i < 8; i++)
  {
    tft.drawNumber(espNowData32BitArray[i], espNowByte0BitsXpos[i], espNowByte0BitsYpos[i]);
    tft.drawNumber(espNowData32BitArray[i + 8], espNowByte1BitsXpos[i], espNowByte1BitsYpos[i]);
    tft.drawNumber(espNowData32BitArray[i + 16], espNowByte2BitsXpos[i], espNowByte2BitsYpos[i]);
    tft.drawNumber(espNowData32BitArray[i + 24], espNowByte3BitsXpos[i], espNowByte3BitsYpos[i]);
  }
}


// **************************************************************************************
// Screen - create bit array from uint32_t to transmit
void BitsArray()
{
  z = 31;
  for (i = 0; i < 32; i++)
  {
    espNowData32BitArray[z] = 0 != (x & (1 << i));
    z--;
  }
}


// **************************************************************************************
void ESPNowSetup()
{
  // ESP-NOW - set device as a Wi-Fi Station
  WiFi.mode(WIFI_STA);

  // ESP-NOW - init
  if (esp_now_init() != ESP_OK)
  {
    tft.fillCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0xF800);
    tft.drawCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0x9492);
  }
  else
  {
    tft.fillCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0x0700);
    tft.drawCircle(espNowStatusIconXpos[0], espNowStatusIconYpos[0], espNowStatusIconRadius[0], 0x9492);
  }

  // ESP-NOW - register for send CB to get the status of trasnmitted packet
  esp_now_register_send_cb(ESPNowOnSend);

  // ESP-NOW - register peer MAC address
  memcpy(peerInfo.peer_addr, broadcastAddress, 6);
  peerInfo.channel = 0;
  peerInfo.encrypt = false;

  // ESP-NOW - add peer
  if (esp_now_add_peer(&peerInfo) != ESP_OK)
  {
    tft.fillCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0xF800);
    tft.drawCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0x9492);
  }
  else
  {
    tft.fillCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0x0700);
    tft.drawCircle(espNowStatusIconXpos[1], espNowStatusIconYpos[1], espNowStatusIconRadius[1], 0x9492);
  }
}


// **************************************************************************************
void ESPNowSend()
{
  // ESP-NOW - reset send and delivery status
  tft.fillCircle(espNowStatusIconXpos[2], espNowStatusIconYpos[2], espNowStatusIconRadius[2] + 1, 0x0000);
  tft.fillCircle(espNowStatusIconXpos[3], espNowStatusIconYpos[3], espNowStatusIconRadius[3] + 1, 0x0000);

  // ESP-NOW - uint32_t to send
  espNowData.espNowData32Bit = x;

  // ESP Now - send message and display error / success
  //esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &espNowData, sizeof(espNowData));
  if (esp_now_send(broadcastAddress, (uint8_t *) &espNowData, sizeof(espNowData)) != ESP_OK)
  {
    tft.fillCircle(espNowStatusIconXpos[2], espNowStatusIconYpos[2], espNowStatusIconRadius[2], 0xF800);
    tft.drawCircle(espNowStatusIconXpos[2], espNowStatusIconYpos[2], espNowStatusIconRadius[2], 0x9492);
  }
  else
  {
    tft.fillCircle(espNowStatusIconXpos[2], espNowStatusIconYpos[2], espNowStatusIconRadius[2], 0x0700);
    tft.drawCircle(espNowStatusIconXpos[2], espNowStatusIconYpos[2], espNowStatusIconRadius[2], 0x9492);
  }

  // ESP Now - time for the receiver to proceed a message
  delay(espNowMsgDelay);
}
